#include <stdio.h>
#include <string.h>
#include "LoadFlat.h"

//#define DBG
#define REVERSE4(x) (((x & 0xFF000000)>>24) | ((x & 0x00FF0000)>>8) | ((x & 0x0000FF00)<<8) | ((x & 0x000000FF)<<24))

FLT_RESULT _FLT_CheckHeader ( flat_hdr *pHdr )
{
	/* Check the magic field */
	
	if ( ! ( 'b' == pHdr->magic[ 0 ] &&
			 'F' == pHdr->magic[ 1 ] &&
			 'L' == pHdr->magic[ 2 ] &&
			 'T' == pHdr->magic[ 3 ] ) )
		return FLT_ERROR_PARSEERROR;
		
	if ( 5 != REVERSE4 ( pHdr->rev ) )
		return FLT_ERROR_PARSEERROR;
		
	return FLT_ERROR_NOERROR;
}

FLT_RESULT FLT_GetSizes	(
							const FLT_FILE pInFlat,
							unsigned long *pdwSizeText,
							unsigned long *pdwSizeData,
							unsigned long *pdwSizeBSS
						)
{
	FLT_RESULT result;
	flat_hdr *pHdr = ( flat_hdr * ) pInFlat;
	
	if ( result = _FLT_CheckHeader ( pHdr ) )
		return result;
		
	*pdwSizeText = REVERSE4(pHdr->data_start) - REVERSE4(pHdr->entry     );
	*pdwSizeData = REVERSE4(pHdr->data_end  ) - REVERSE4(pHdr->data_start);
	*pdwSizeBSS  = REVERSE4(pHdr->bss_end   ) - REVERSE4(pHdr->data_end  );

	return FLT_ERROR_NOERROR;
}

FLT_RESULT _FLT_RelocateFlat (
							unsigned char *pFlat,
							FLT_SECTION pInText,
							FLT_SECTION pInData,
							const FLT_SECTION pInBSS,
							void **ppText
						)
{
	flat_hdr *pHdr = ( flat_hdr * ) pFlat;
	flat_v5_reloc_t *pRelocTable = (flat_v5_reloc_t *) ( pFlat + REVERSE4(pHdr->reloc_start) );

	unsigned int i;
	FLT_RESULT result;
	
#ifdef DBG
	printf ( "Rev %d\n", REVERSE4(pHdr->rev ) );
	printf ( "Entry 0x%x [0x%x]\n", REVERSE4(pHdr->entry), REVERSE4(pHdr->entry) + pFlat );
	printf ( "Data_Start 0x%x\n", REVERSE4(pHdr->data_start) );
	printf ( "Data_End 0x%x\n", REVERSE4(pHdr->data_end) );
	printf ( "BSS_End 0x%x\n", REVERSE4(pHdr->bss_end) );
	printf ( "Stack_Size 0x%x\n", REVERSE4(pHdr->stack_size) );
	printf ( "Reloc_Start 0x%x [0x%x]\n", REVERSE4(pHdr->reloc_start), REVERSE4(pHdr->reloc_start) +pFlat );
	printf ( "Reloc_Count %d\n", REVERSE4(pHdr->reloc_count) );
	printf ( "Flags %d\n", REVERSE4(pHdr->flags) );
	printf ( "Stamp 0x%x\n", REVERSE4(pHdr->build_date) );
	printf ( "<eof> 0x%x\n", REVERSE4(pHdr->reloc_start) + REVERSE4(pHdr->reloc_count)*4);
	printf ( "%4s %2s %2s %6s %6s %8s\n", "type", "hi", "sp", "offset", "addr", "value" );
#endif
	for ( i = 0; i < REVERSE4(pHdr->reloc_count); ++i)
	{
		unsigned char *pAddressToBePatched;
		unsigned long pAddressPatchValue;
		flat_v5_reloc_t reloc;
		
		// Using memcpy() here rather than a simple assignment since pRelocTable may
		// have a value that is not 4-byte aligned.
		
		memcpy ( &reloc.value, &((pRelocTable+i)->value), sizeof ( unsigned long ) );
		reloc.value = REVERSE4( reloc.value );
			
		switch ( reloc.reloc.type )
		{
			case FLAT_RELOC_TYPE_TEXT:
				
				pAddressPatchValue = (unsigned long) pInText;
				break;
				
			case FLAT_RELOC_TYPE_DATA:

				pAddressPatchValue = (unsigned long) pInData;
				break;
				
			case FLAT_RELOC_TYPE_BSS:
			
				pAddressPatchValue = (unsigned long) pInBSS;
				break;
				
			default:
			
				return FLT_ERROR_PARSEERROR;
		}
		
#ifdef DBG
		{
			char *sec;
			int width;

			switch (reloc.reloc.type)
			{
				case 0: sec = "text";
					break;
				case 1: sec = "data";
					break;
				case 2: sec = "bss ";
					break;
				default: sec = "????";
			}

			switch (reloc.reloc.sp)
			{
				case 0: width=16;
					break;
				case 1: width=24;
					break;
				case 2: width=32;
					break;
				default: width=-1;
			}

		printf ( "%4s %2d %2d 0x%-6x 0x%-7x 0x%6x\n", sec, reloc.reloc.hi_lo, width, reloc.reloc.offset, reloc.reloc.offset+REVERSE4(pHdr->entry), reloc.value);
		}
#endif
		unsigned long fltTextSize = REVERSE4(pHdr->data_start) - REVERSE4(pHdr->entry);
		unsigned long fltDataSize = REVERSE4(pHdr->data_end) - REVERSE4(pHdr->data_start);
		if ( reloc.reloc.offset < fltTextSize )
			pAddressToBePatched = pInText + reloc.reloc.offset;
		else if ( reloc.reloc.offset < (fltTextSize + fltDataSize))
			pAddressToBePatched = pInData + reloc.reloc.offset - fltTextSize;
		else 
			return FLT_ERROR_PARSEERROR;

		unsigned short *p16 = (unsigned short *) pAddressToBePatched;
		unsigned long  *p32 = (unsigned long  *) pAddressToBePatched;
		
		switch ( reloc.reloc.sp )
		{
			
			case 0:
			
				p16++; // Increment to the second half of the instruction
				if ( reloc.reloc.hi_lo )
				{
					// Upper 16 bits
					*p16 += ( pAddressPatchValue >> 16 ) & 0xffff;
				} else {
					// Lower 16 bits
					*p16 += pAddressPatchValue & 0xffff;
				}
				break;
			
			case 1:

				p16++; // Increment to the second half of the instruction
                unsigned long PatchValueAddend = 0;
				memcpy ( &PatchValueAddend, p16, sizeof ( unsigned short ) );
                
				if(PatchValueAddend == 0) {
					memcpy ( &PatchValueAddend, &((pRelocTable+(++i))->value), sizeof ( unsigned long ) );
				}else{				                
					memcpy ( &PatchValueAddend, &((pRelocTable+PatchValueAddend)->value), sizeof ( unsigned long ) );
				}
                
                pAddressPatchValue += REVERSE4( PatchValueAddend );
				if ( reloc.reloc.hi_lo )
				{	// Upper 16 bits
					*p16 = ( pAddressPatchValue >> 16 ) & 0xffff;
				} else 
				{	// Lower 16 bits
					*p16 = pAddressPatchValue & 0xffff;
				}

				break;

			case 2:
			
				*p32 += (unsigned long) pAddressPatchValue;
				break;
					
			default:
			
				return FLT_ERROR_PARSEERROR;
		}
	}
	
	*ppText = pInText;
	
	return FLT_ERROR_NOERROR;
}

FLT_RESULT FLT_RelocateFlat	(
							const FLT_FILE pInFlat,
							FLT_SECTION pInText,
							FLT_SECTION pInData,
							const FLT_SECTION pInBSS,
							void **ppText
						)
{
	unsigned char *pFlat = pInFlat;
	flat_hdr *pHdr = ( flat_hdr * ) pFlat;
	FLT_RESULT result;

	if ( result = _FLT_CheckHeader ( pHdr ) )
		return result;
		
	memcpy ( pInText, pFlat + REVERSE4(pHdr->entry     ), REVERSE4(pHdr->data_start) - REVERSE4(pHdr->entry     ) );
	memcpy ( pInData, pFlat + REVERSE4(pHdr->data_start), REVERSE4(pHdr->data_end  ) - REVERSE4(pHdr->data_start) );
		
	return _FLT_RelocateFlat ( pFlat, pInText, pInData, pInBSS, ppText );
}

FLT_RESULT FLT_RelocateFlatInPlace	(
							FLT_FILE pInFlat,
							const FLT_SECTION pInBSS,
							void **ppText
						)
{
	unsigned char *pFlat = pInFlat;
	flat_hdr *pHdr = ( flat_hdr * ) pFlat;
	FLT_RESULT result;
	
	if ( result = _FLT_CheckHeader ( pHdr ) )
		return result;
		
	return _FLT_RelocateFlat (	pFlat,
								pFlat + REVERSE4(pHdr->entry),
								pFlat + REVERSE4(pHdr->data_start),
								pInBSS,
								ppText );
}

#include "export.h"

FLT_RESULT FLT_LookupByName ( const FLT_SECTION pText, const char *pName, void **ppSymbol )
{
	struct _exported_symbol *pExportedSymbols;
	pExportedSymbols = ( struct _exported_symbol * ) pText;
	
	for ( ; pExportedSymbols->sym; ++pExportedSymbols )
	{
		if ( !strcmp ( pExportedSymbols->sym, pName ) )
		{
			*ppSymbol = pExportedSymbols->ptr;
			return FLT_ERROR_NOERROR;
		}
	}

	return FLT_ERROR_NAMENOTFOUND;
}



