#ifndef _LOADFLAT_H_
#define _LOADFLAT_H_

/* FLAT data structures derived from uClinux sources */

#define __LITTLE_ENDIAN_BITFIELD

typedef union
{
  unsigned long	value;
  struct {
#if defined(__LITTLE_ENDIAN_BITFIELD)
    unsigned long type  : 2;
    unsigned long hi_lo : 1;
    unsigned long sp    : 3;
    signed long offset  : 26;
#elif defined(__BIG_ENDIAN_BITFIELD)
    signed long offset  : 26;
    unsigned long sp    : 3;
    unsigned long hi_lo : 1;
    unsigned long type  : 2;
#else
#error "Unknown bitfield order for flat files."
#endif
	} reloc;
} flat_v5_reloc_t;

#define FLAT_RELOC_TYPE_TEXT 0
#define FLAT_RELOC_TYPE_DATA 1
#define FLAT_RELOC_TYPE_BSS 2
#define FLAT_RELOC_PART_LO 0
#define FLAT_RELOC_PART_HI 1

typedef struct {
	char magic[4];
	unsigned long rev;          /* version (as above) */
	unsigned long entry;        /* Offset of first executable instruction
	                               with text segment from beginning of file */
	unsigned long data_start;   /* Offset of data segment from beginning of
	                               file */
	unsigned long data_end;     /* Offset of end of data segment
	                               from beginning of file */
	unsigned long bss_end;      /* Offset of end of bss segment from beginning
	                               of file */

	/* (It is assumed that data_end through bss_end forms the bss segment.) */

	unsigned long stack_size;   /* Size of stack, in bytes */
	unsigned long reloc_start;  /* Offset of relocation records from
	                               beginning of file */
	unsigned long reloc_count;  /* Number of relocation records */
	unsigned long flags;       
        unsigned long build_date;  /* bfin changes added build_date */     
	unsigned long filler[6];    /* Reservered, set to zero */
} flat_hdr;

/* ADI-developed content begins here. */

/* Flat File API return codes */

typedef unsigned char *  FLT_FILE    ;
typedef unsigned char *  FLT_SECTION ;

typedef enum
{
	FLT_ERROR_NOERROR		= 0,
	FLT_ERROR_UNIMPLEMENTED	= 98,
	FLT_ERROR_UNKNOWN		= 99,
	FLT_ERROR_OUTOFMEMORY	= 100,
	FLT_ERROR_PARSEERROR,
	FLT_ERROR_NAMENOTFOUND,
} FLT_RESULT;

/* Flat File API */

FLT_RESULT FLT_GetSizes ( const FLT_FILE, unsigned long *, unsigned long *, unsigned long *);
FLT_RESULT FLT_RelocateFlat ( const FLT_FILE, FLT_SECTION, FLT_SECTION, const FLT_SECTION, void ** );
FLT_RESULT FLT_RelocateFlatInPlace ( FLT_FILE, const FLT_SECTION, void ** );
FLT_RESULT FLT_LookupByName ( const FLT_SECTION, const char *, void ** );

#endif

